signature CANON =  sig

  datatype canon =
    Ws  of (Context.ctxt * string)
  | Txt of (Context.ctxt * string)
  | Seq of canon * canon

  val canon   : SMLNG.doc -> canon
  val toDoc   : canon -> SMLNG.doc

  val jlist   : SMLNG.doc -> (Context.ctxt * string) JoinList.jlist
  val sameDoc : SMLNG.doc * SMLNG.doc -> bool 

end
structure Canon :> CANON =
  struct
  datatype canon = 
    Txt of (Context.ctxt * string)
  | Ws  of (Context.ctxt * string)
  | Seq of canon * canon

  fun canon doc =
    Context.fold_with_ctxt {kSeq=Seq,
			    kTxt=Txt,
			    kWs=Ws,
			    kNd=(fn (_,t,a) => a)} Context.init doc

  fun mkNode f [] t = f t
    | mkNode f (x::xs) t = SMLNG.Nd(x,mkNode f xs t)

  fun toDoc (Txt(ctxt,text)) = mkNode SMLNG.Txt (Context.toTags ctxt) text
    | toDoc (Ws(ctxt,text)) = mkNode SMLNG.Ws (Context.toTags ctxt) text
    | toDoc (Seq(d1,d2)) = SMLNG.Seq(toDoc d1,toDoc d2)

  fun jlist doc =
    Context.fold_with_ctxt {kSeq=JoinList.append,
			    kTxt=JoinList.one,
			    kWs=JoinList.one,
			    kNd=(fn (_,t,a) => a)} Context.init doc

  fun sameDoc (d1,d2) = let
    val d1 = jlist d1
    val d2 = jlist d2
    fun eq (JoinList.Nil,JoinList.Nil) = true
      | eq (JoinList.Cons((cx,sx:string),xs),JoinList.Cons((cy,sy),ys)) =
      Context.eq (cx,cy) andalso 
      (sx = sy) andalso eq(JoinList.pat xs,JoinList.pat ys)
      | eq (xs,ys) = false
  in  eq(JoinList.pat d1,JoinList.pat d2)
  end
  end