structure Main =
  struct
    fun parse  s = let
      val ins = TextIO.openIn s
      val doc = SMLNG.parse ins
    in TextIO.closeIn ins;doc
    end

    exception Bigger
    exception NotSame
    val dbg = ref true
    fun dsay s = if !dbg then TextIO.output(TextIO.stdErr,s) else ()
    fun check n f old = let
      val new = f old
      val old_sz = SMLNG.size old
      val new_sz = SMLNG.size new
      val same = Canon.sameDoc (new,old)
      val bigger = new_sz > old_sz
      val _ = dsay ("# pass: "^n^"\n");
      val _ = dsay ("# old size: "^(Int.toString old_sz)^"\n");
      val _ = dsay ("# new size: "^(Int.toString new_sz)^"\n");
    in (if bigger then raise Bigger
       else if not same then raise NotSame
	    else new) handle e =>
      (dsay ("# err: "^n^"\n"); 
       if !dbg then SMLNG.print new TextIO.stdErr else ();
	 raise e)
    end

    fun test f s = let
      val doc = parse s
      val _ = print "# input\n"
      val _ = SMLNG.print doc TextIO.stdOut
      val x = f doc
      val _ = print "\n# output\n"
      val _ = SMLNG.print x TextIO.stdOut
      val _ = print "\n# eof\n"
    in x
    end

    val canon = test (Canon.toDoc o Canon.canon)
    val cdoc  = test (check "cdoc" (CDocs.toDoc o CDocs.fromDoc))
    val em_elim = test (check "em-elim" EMElimination.go)
    val inter_ng = test (check "inter-ng" InterNG.go)
    val redundant = test (check "redundant" RemoveRedundant.go)
   val merge_adj  = test (check "merge-adj" (MergeAdjacent.go)) 

    local 
      val last_good = ref (SMLNG.Txt "")
    in fun get_good () = !last_good
       fun do_pass n f x = let
	 val _ = last_good := x
	 val x = check n f x
       in last_good := x;x
       end
       fun try_pass n f x = let
	 val _ = last_good := x
	 val x = (check n f x) handle _ => x
       in last_good := x;x
       end
    end
    val bogus = (fn x => raise (Fail ""))
    val opt = List.foldr (fn ((n,f),g) => (g o (try_pass n f))) (fn x => x) 
      [("merge-adj",MergeAdjacent.go),
       ("redundant",RemoveRedundant.go),
       ("em-elim", EMElimination.go),
       ("inter_ng", InterNG.go),
       ("merge-adj",MergeAdjacent.go),
       ("redundant",RemoveRedundant.go),
       ("em-elim", EMElimination.go)]

    fun fail_safe t f x = 
      ((TimeLimit.timeLimit (Time.fromSeconds t) f x) handle _ => 
	(dsay("\n# Fail safe invoked\n"); get_good()))
    fun toInt dflt s = Option.getOpt(LargeInt.fromString s,dflt)
    val dflt_tlimit = LargeInt.fromInt 180
    fun main (_,args) = let
      val (tlimit,doc) = 
	case args of [x] =>
	  (toInt dflt_tlimit x,SMLNG.parse  TextIO.stdIn)
      | (x::y::_) => (toInt dflt_tlimit x,parse y)
      | [] => (dflt_tlimit,SMLNG.parse  TextIO.stdIn)
      val slop = 20 (* twenty seconds of slack time *)
      val doc = fail_safe (tlimit - slop) opt doc
    in SMLNG.print doc TextIO.stdOut;
       OS.Process.success
    end
  end