structure SMLNG :> SML_NG  =
  struct
    datatype doc =
      Seq of doc * doc
    | Txt of string
    | Ws  of string
    | Nd of tag * doc
    and tag =
        B
      | EM
      | I
      | PL
      | S
      | TT
      | U
      | Size  of int
      | Color of color 
  and color =  Red | Green | Blue | Cyan | Magenta | Yellow | Black | White

    fun tag2str B = "B"
      | tag2str EM = "EM"
      | tag2str I = "I"
      | tag2str PL = "PL"
      | tag2str S = "S"
      | tag2str TT = "TT"
      | tag2str U = "U"
      | tag2str (Size 0) = "0"
      | tag2str (Size 1) = "1"
      | tag2str (Size 2) = "2"
      | tag2str (Size 3) = "3"
      | tag2str (Size 4) = "4"
      | tag2str (Size 5) = "5"
      | tag2str (Size 6) = "6"
      | tag2str (Size 7) = "7"
      | tag2str (Size 8) = "8"
      | tag2str (Size 9) = "9"
      | tag2str (Color Red) = "r"
      | tag2str (Color Green) = "g"
      | tag2str (Color Blue) = "b"
      | tag2str (Color Cyan) = "c"
      | tag2str (Color Magenta) = "m"
      | tag2str (Color Yellow) = "y"
      | tag2str (Color Black) = "k"
      | tag2str (Color White) = "w"
      | tag2str (Size _) = raise (Fail "invalid size")

    fun str2tag "B" = B 
      | str2tag "EM" = EM 
      | str2tag "I" = I 
      | str2tag "PL" = PL 
      | str2tag "S" = S 
      | str2tag "TT" = TT 
      | str2tag "U" = U 
      | str2tag "0" = (Size 0) 
      | str2tag "1" = (Size 1) 
      | str2tag "2" = (Size 2) 
      | str2tag "3" = (Size 3) 
      | str2tag "4" = (Size 4) 
      | str2tag "5" = (Size 5) 
      | str2tag "6" = (Size 6) 
      | str2tag "7" = (Size 7) 
      | str2tag "8" = (Size 8) 
      | str2tag "9" = (Size 9) 
      | str2tag "r" = (Color Red) 
      | str2tag "g" = (Color Green) 
      | str2tag "b" = (Color Blue) 
      | str2tag "c" = (Color Cyan) 
      | str2tag "m" = (Color Magenta) 
      | str2tag "y" = (Color Yellow) 
      | str2tag "k" = (Color Black) 
      | str2tag "w" = (Color White) 
      | str2tag s = raise (Fail ("bad tag:"^s))

    fun chars2tag [#"B"] = B 
      | chars2tag [#"E",#"M"] = EM 
      | chars2tag [#"I"] = I 
      | chars2tag [#"P",#"L"] = PL 
      | chars2tag [#"S"] = S 
      | chars2tag [#"T",#"T"] = TT 
      | chars2tag [#"U"] = U 
      | chars2tag [#"0"] = (Size 0) 
      | chars2tag [#"1"] = (Size 1) 
      | chars2tag [#"2"] = (Size 2) 
      | chars2tag [#"3"] = (Size 3) 
      | chars2tag [#"4"] = (Size 4) 
      | chars2tag [#"5"] = (Size 5) 
      | chars2tag [#"6"] = (Size 6) 
      | chars2tag [#"7"] = (Size 7) 
      | chars2tag [#"8"] = (Size 8) 
      | chars2tag [#"9"] = (Size 9) 
      | chars2tag [#"r"] = (Color Red) 
      | chars2tag [#"g"] = (Color Green) 
      | chars2tag [#"b"] = (Color Blue) 
      | chars2tag [#"c"] = (Color Cyan) 
      | chars2tag [#"m"] = (Color Magenta) 
      | chars2tag [#"y"] = (Color Yellow) 
      | chars2tag [#"k"] = (Color Black) 
      | chars2tag [#"w"] = (Color White) 
      | chars2tag cs = raise (Fail ("bad tag:"^(String.implode cs)))



    datatype tok = 
        STR  of string
      | WS   of string
      | BTAG of tag
      | ETAG of tag

    fun isWs #" " = true
      | isWs #"\n" = true
      | isWs #"\r" = true
      | isWs #"\t" = true
      | isWs #"\f" = true
      | isWs _ = false

    fun take_until_gt getch strm = let
      fun loop strm = 
	case getch strm of
	  SOME (c,strm) =>
	    if c = #">" then SOME ([],strm)
	    else (case (loop strm) of
		    SOME (cs,strm) => SOME (c::cs,strm)
		  | NONE => NONE)
	| NONE => NONE
    in loop strm
    end


    fun get_ws getch strm = let
      fun loop strm = 
	case getch strm of
	  SOME (c,strm') =>
	    if not (isWs c) then SOME ([],strm)
	    else (case (loop strm) of
		    SOME (cs,strm) => SOME (c::cs,strm)
		  | NONE => NONE)
	| NONE => NONE
    in loop strm
    end

    fun eat_ws getch strm = let
      fun loop strm = 
	case getch strm of
	  SOME (c,strm') => if isWs c then loop(strm')
			    else strm
	| NONE => strm
    in loop strm
    end

    fun get_word getch strm = let
      fun loop strm = 
	case getch strm of
	  SOME (c,strm') =>
	    if (isWs c) orelse (c = #"<") then SOME ([],strm)
	    else (case (loop strm') of
		    SOME (cs,strm) => SOME (c::cs,strm)
		  | NONE => NONE)
	| NONE => NONE
    in loop strm
    end

    fun get_text true getch strm = let
      fun loop (strm,acc) = 
 	case getch strm of
 	  SOME (c,strm') =>
 	    if c = #"<" then SOME (List.rev acc,strm)
 	    else loop(strm',c::acc)
 	| NONE => if List.null acc then NONE
 		  else (SOME (List.rev acc,strm))
    in loop(strm,[])
    end
      | get_text false getch strm = let
	  fun loop (strm,acc) = 
	    case getch strm of
	      SOME (c,strm') =>
		if c = #"<" then SOME (List.rev acc,strm)
		else if isWs(c) then loop(eat_ws getch strm',#" "::acc)
		     else loop(strm',c::acc)
	    | NONE => if List.null acc then NONE
		      else (SOME (List.rev acc,strm))
	in loop(strm,[])
	end

(*      
    fun parse_tok getch keep_ws strm =
      (case (getch strm) of
	 SOME (#"<",strm) => let
	   val (mktok,strm) =
	     case (getch strm) of
	      SOME (#"/",strm) => (ETAG,strm)
	    | SOME _  => (BTAG,strm)
	    | _ => raise (Fail "incomplete tag")
	   val (cs,strm) =
	     case take_until_gt getch strm of
	       SOME x => x
	     | NONE => raise (Fail "missing >")
	   val tag = chars2tag cs
	 in SOME (mktok tag,strm)
	 end
       | SOME (c,_) => 
	  if isWs(c) then
	    if keep_ws then
	      (case (get_ws getch strm) of
		SOME(cs,strm) => SOME(WS (String.implode cs),strm)
	      | NONE => NONE)
	    else SOME(WS " ",eat_ws getch strm)
	  else (case get_word getch strm of
		  SOME (cs,strm) => SOME (STR (String.implode cs),strm)
		| NONE => NONE)
       | NONE => NONE)
*)
    fun parse_tok getch keep_ws strm =
      (case (getch strm) of
	 SOME (#"<",strm) => let
	   val (mktok,strm) =
	     case (getch strm) of
	      SOME (#"/",strm) => (ETAG,strm)
	    | SOME _  => (BTAG,strm)
	    | _ => raise (Fail "incomplete tag")
	   val (cs,strm) =
	     case take_until_gt getch strm of
	       SOME x => x
	     | NONE => raise (Fail "missing >")
	   val tag = chars2tag cs
	 in SOME (mktok tag,strm)
	 end
       | SOME (c,_) =>
	 (case get_text keep_ws getch strm of
	    SOME(cs,strm) =>
	      if List.all isWs cs then SOME(WS (String.implode cs),strm)
	      else SOME(STR (String.implode cs),strm)
	 | NONE => NONE)
       | NONE => NONE)
    fun list_seq [] = Txt ""
      | list_seq [t] = t
      | list_seq (x::xs) = Seq(x,list_seq xs)

    fun parse_docs (gettok,keep_ws) strm =
      case gettok keep_ws strm of
	SOME(STR s,strm) => let
	  val (docs,strm) = parse_docs (gettok,keep_ws) strm
	in ((Txt s)::docs,strm)
	end
      | SOME(WS s,strm) => let
	  val (docs,strm) = parse_docs (gettok,keep_ws) strm
	in ((Ws s)::docs,strm)
	end
      | SOME(BTAG TT,strm) => let
	  val (docs,strm) = parse_docs (gettok,true) strm
	  val doc = Nd(TT,list_seq docs)
	  val (docs,strm) = parse_docs (gettok,keep_ws) strm
	in (doc::docs,strm)
	end
      | SOME(BTAG t,strm) => let
	  val (docs,strm) = parse_docs (gettok,keep_ws) strm
	  val doc = Nd(t,list_seq docs)
	  val (docs,strm) = parse_docs (gettok,keep_ws) strm
	in (doc::docs,strm)
	end
      | SOME(ETAG t,strm) => ([],strm)
      | NONE => ([],strm)

    fun parse ins = let
      fun parse_doc getch strm = let
	val gettok = parse_tok getch 
	val keep_ws = false
	fun loop ([],strm) = ([],strm)
	  | loop (docs,strm) = let
	      val (docss,strm) = loop (parse_docs (gettok,keep_ws) strm)
	    in ((list_seq docs)::docss,strm)
	    end
      in loop (parse_docs (gettok,keep_ws) strm)

      end

      fun dbgetch s =
	case TextIO.StreamIO.input1 s of
	NONE => (print ("EOF"); NONE)
      | SOME(c,strm) => (print (Char.toString c);SOME(c,strm))
      val strm = TextIO.getInstream ins
      val (docs,strm) = (parse_doc TextIO.StreamIO.input1 strm) 
(*      val (docs,strm) = (parse_doc dbgetch strm) *)
    in TextIO.setInstream (ins,strm);list_seq docs
    end
  
    fun print doc out = let
      fun pr s = TextIO.output(out,s)
      fun f (Seq(d1,d2)) = (f d1;f d2)
	| f (Txt s) = pr s
	| f (Ws s) = pr s
	| f (Nd (t,d)) = let
	    val t = tag2str t
	  in pr ("<");pr t;pr(">");
	    f d;
	    pr ("</");pr t;pr(">")
	  end
    in f doc
    end
     fun fold {kSeq,kTxt,kWs,kNd} d = let
       fun f (Seq(d1,d2)) = kSeq(f d1,f d2)
	 | f (Txt s) = kTxt s
	 | f (Ws s) = kWs s
	 | f (Nd (t,d)) = kNd(t,f d)
     in f d
     end
    fun tagsize EM = 9
      | tagsize PL = 9
      | tagsize TT = 9
      | tagsize _ = 7

     fun size d =
       fold {kSeq=(fn (x,y) => x + y),
	     kTxt=(fn s => String.size s),
	     kWs=(fn s => String.size s),
	     kNd=(fn(t,x) => (tagsize t) + x)} d
  end
